@@ -1,5 +1,5 @@
 /* copy.c -- core functions for copying files and directories
-   Copyright (C) 89, 90, 91, 1995-2006 Free Software Foundation.
+   Copyright (C) 89, 90, 91, 1995-2007 Free Software Foundation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -295,8 +295,8 @@ copy_reg (char const *src_name, char const *dst_name,
       goto close_src_desc;
     }
 
-  /* These semantics are required for cp.
-     The if-block will be taken in move_mode.  */
+  /* The semantics of the following open calls are mandated
+     by the specs for both cp and mv.  */
   if (! *new_dst)
     {
       dest_desc = open (dst_name, O_WRONLY | O_TRUNC | O_BINARY);
@@ -430,7 +430,17 @@ copy_reg (char const *src_name, char const *dst_name,
 	    {
 	      char *cp;
 
-	      buf[n_read] = 1;	/* Sentinel to stop loop.  */
+	      /* Sentinel to stop loop.  */
+	      buf[n_read] = '\1';
+#ifdef lint
+	      /* Usually, buf[n_read] is not the byte just before a "word"
+		 (aka uintptr_t) boundary.  In that case, the word-oriented
+		 test below (*wp++ == 0) would read some uninitialized bytes
+		 after the sentinel.  To avoid false-positive reports about
+		 this condition (e.g., from a tool like valgrind), set the
+		 remaining bytes -- to any value.  */
+	      memset (buf + n_read + 1, 0, sizeof (word) - 1);
+#endif
 
 	      /* Find first nonzero *word*, or the word with the sentinel.  */
 
@@ -1005,6 +1015,7 @@ copy_internal (char const *src_name, char const *dst_name,
   struct stat dst_sb;
   mode_t src_mode;
   mode_t dst_mode IF_LINT (= 0);
+  mode_t dst_mode_bits;
   mode_t omitted_permissions;
   bool restore_dst_mode = false;
   char *earlier_file = NULL;
@@ -1504,13 +1515,16 @@ copy_internal (char const *src_name, char const *dst_name,
       new_dst = true;
     }
 
-  /* If the ownership might change, omit some permissions at first, so
-     unauthorized users cannot nip in before the file has the right
-     ownership.  */
+  /* If the ownership might change, or if it is a directory (whose
+     special mode bits may change after the directory is created),
+     omit some permissions at first, so unauthorized users cannot nip
+     in before the file is ready.  */
+  dst_mode_bits = (x->set_mode ? x->mode : src_mode) & CHMOD_MODE_BITS;
   omitted_permissions =
-    (x->preserve_ownership
-     ? (x->set_mode ? x->mode : src_mode) & (S_IRWXG | S_IRWXO)
-     : 0);
+    (dst_mode_bits
+     & (x->preserve_ownership ? S_IRWXG | S_IRWXO
+	: S_ISDIR (src_mode) ? S_IWGRP | S_IWOTH
+	: 0));
 
   delayed_ok = true;
 
@@ -1549,10 +1563,7 @@ copy_internal (char const *src_name, char const *dst_name,
 	     (src_mode & ~S_IRWXUGO) != 0.  However, common practice is
 	     to ask mkdir to copy all the CHMOD_MODE_BITS, letting mkdir
 	     decide what to do with S_ISUID | S_ISGID | S_ISVTX.  */
-	  mode_t mkdir_mode =
-	    ((x->set_mode ? x->mode : src_mode)
-	     & CHMOD_MODE_BITS & ~omitted_permissions);
-	  if (mkdir (dst_name, mkdir_mode) != 0)
+	  if (mkdir (dst_name, dst_mode_bits & ~omitted_permissions) != 0)
 	    {
 	      error (0, errno, _("cannot create directory %s"),
 		     quote (dst_name));
@@ -1686,12 +1697,16 @@ copy_internal (char const *src_name, char const *dst_name,
     {
       /* Use mknod, rather than mkfifo, because the former preserves
 	 the special mode bits of a fifo on Solaris 10, while mkfifo
-	 does not.  */
+	 does not.  But fall back on mkfifo, because on some BSD systems,
+	 mknod always fails when asked to create a FIFO.  */
       if (mknod (dst_name, src_mode & ~omitted_permissions, 0) != 0)
-	{
-	  error (0, errno, _("cannot create fifo %s"), quote (dst_name));
-	  goto un_backup;
-	}
+#if HAVE_MKFIFO
+	if (mkfifo (dst_name, src_mode & ~S_IFIFO & ~omitted_permissions) != 0)
+#endif
+	  {
+	    error (0, errno, _("cannot create fifo %s"), quote (dst_name));
+	    goto un_backup;
+	  }
     }
   else if (S_ISBLK (src_mode) || S_ISCHR (src_mode) || S_ISSOCK (src_mode))
     {
@@ -1975,7 +1990,7 @@ chown_failure_ok (struct cp_options const *x)
 extern mode_t
 cached_umask (void)
 {
-  static mode_t mask = -1;
+  static mode_t mask = (mode_t) -1;
   if (mask == (mode_t) -1)
     {
       mask = umask (0);
